@@ -1,4 +1,5 @@
-sp.polygons = function(obj, col = 1, fill="transparent", ...) {
+setMethod("sppanel", "SpatialPolygons",
+  function(obj, col = 1, fill = "transparent", ...) {
 	if (is.character(obj))
 		obj = get(obj)
 	if (!is(obj, "SpatialPolygons"))
@@ -43,33 +44,37 @@ sp.polygons = function(obj, col = 1, fill="transparent", ...) {
 				sp.polygon3(Srs[[j]], col = col, fill = fill[i], ...)
 		}
 	}
-}
+})
 
-sp.lines = function(obj, col = 1, ...) {
-	if (is.character(obj))
-		obj = get(obj)
-	sp.lines3 = function(x, col, ...) panel.lines(coordinates(x), col = col, ...)
-	sp.lines2 = function(x, col, ...) lapply(x@Lines, sp.lines3, col = col, ...)
-	if (is(obj, "SpatialLines"))
-		lapply(obj@lines, sp.lines2, col = col, ...)
-	else if (is(obj, "Lines"))
-		lapply(obj@Lines, sp.lines3, col = col, ...)
-	else if (is(obj, "Line"))
-		panel.lines(coordinates(obj), col = col, ...)
-	else stop(paste("obj of class Line, Lines or SpatialLines expected, got", class(obj)))
-}
+# backward compatibility:
+sp.polygons = function(obj, col = 1, fill = "transparent",...) 
+	sppanel(obj, col = col, fill = fill, ...)
 
-sp.points = function(obj, pch = 3, ...) {
-	if (is.character(obj))
-		obj = get(obj)
-	xy = coordinates(obj)
-	panel.points(xy[,1], xy[,2], pch = pch, ...)
-}
+setMethod("sppanel", "SpatialLines", 
+  function (obj, col = 1, ...) {
+    ## contributed by Josh O'Brien, Mar 15, 2015
+	lo <- length(obj@lines)
+	col <- rep(col, length.out = lo)
+	lapply(seq_len(lo), function(ii) sppanel(obj@lines[[ii]], col = col[ii], ...))
+})
+
+setMethod("sppanel", "Lines",
+  function (obj, col = 1, ...) lapply(obj@Lines, sppanel, col = col, ...))
+
+setMethod("sppanel", "Line",
+  function (obj, col = 1, ...) panel.lines(coordinates(obj), col = col, ...))
+
+# backward compatibility:
+sp.lines = function(obj, col = 1,...) sppanel(obj, col = col,...)
+
+setMethod("sppanel", "SpatialPoints",
+	function(obj, pch = 3, ...)
+		panel.points(coordinates(obj), pch = pch, ...))
+# backward compatibility:
+sp.points = function(obj, pch=3, ...) sppanel(obj, pch = pch, ...)
 
 sp.grid = function(obj, col = 1, alpha = 1, ..., at = pretty(obj[[1]]),
 		col.regions = col) {
-	if (is.character(obj))
-		obj = get(obj)
 	xy = coordinates(obj)
 	if (length(col) > 1 && ("data" %in% slotNames(obj))) {
 		z = obj[[1]]
@@ -83,42 +88,34 @@ sp.grid = function(obj, col = 1, alpha = 1, ..., at = pretty(obj[[1]]),
 		height = gt$cellsize[2], default.units = "native",
 		gp = gpar(fill = col, col = NA, alpha = alpha))
 }
+setMethod("sppanel", "SpatialPixels", sp.grid)
+setMethod("sppanel", "SpatialGrid", sp.grid)
 
 sp.text = function(loc, txt, ...) {
 	if (!is.numeric(loc))
-		stop("loc should be numeric")
+		stop("loc (first argument) should be numeric, indicating text locations")
 	if (length(loc) == 2)
 		panel.text(loc[1], loc[2], txt, ...)
 	else if (is.matrix(loc) && ncol(loc) == 2 && nrow(loc) == length(txt))
 		panel.text(loc[,1], loc[,2], txt, ...)
 	else
 		stop("loc and txt have non-matching dimensions")
-	#panel.text(loc[[1]], loc[[2]], txt, ...)
 }
+setMethod("sppanel", "character", function(obj,txt, ...) sp.text(obj, txt, ...))
 
-sp.panel.layout = function(lst, p.number, ...) {
+sp.panel.layout = function(lst, p.number, ...) { # now obsolete...
+	.Deprecated("sppanel")
 	sp.panel0 = function(x, first = FALSE, ...) {
-		if (is.character(x))
-#			obj = get(x)
-			x = get(x)
-		if (!is.null(x$which) && is.na(match(p.number, x$which)))
-			return()
 		if (inherits(x, "list")) {
+			if (!is.null(x$which) && is.na(match(p.number, x$which)))
+				return()
 			# print(paste(class(x), "first val", first, "first obj", x$first))
 			if (!is.null(x$first)) {
 				if (x$first == first)
 					do.call(x[[1]], x[2:length(x)])
 			} else if (!first)
 				do.call(x[[1]], x[2:length(x)])
-		} else if (is(x, "SpatialLines") || is(x, "Lines") || is(x, "Line"))
-			sp.lines(x, ...)
-		else if (is(x, "SpatialPoints"))
-			sp.points(as(x, "SpatialPoints"), ...)
-		else if (is(x, "SpatialPolygons"))
-			sp.polygons(x, ...)
-		else if (is(x, "SpatialPixels") || is(x, "SpatialGrid"))
-			sp.grid(x, ...)
-		else stop(paste("cannot plot object of class", class(x)))
+		} 
 	}
 	if (!is.null(lst$which) && is.na(match(p.number, lst$which)))
 		return()
@@ -135,6 +132,39 @@ sp.panel.layout = function(lst, p.number, ...) {
 		stop(paste("expected object of class list; got object of class", class(lst)))
 }
 
+setMethod("sppanel", "NULL", function(obj,...) { })
+
+sppanelList = function(obj, p.number, first, ...) {
+	missingFirst = missing(first)
+	if (length(obj) == 1 & is.null(obj[[1]]))
+		return()
+	if (!is.null(obj$which) && is.na(match(p.number, obj$which)))
+		return() # false panel
+	else 
+		obj$which = NULL # continue: either all panels, or right panel
+	if (is.list(obj[[1]])) # list-of-lists, recurse:
+		return(lapply(obj, sppanel, p.number = p.number, first = first, ...))
+	opaque = function(x) (is(x, "SpatialPolygons") || is(x, "SpatialGrid") || is(x, "SpatialPixels"))
+	if (is.character(obj[[1]]) || is.function(obj[[1]])) {
+		if (is.null(obj$first))
+			obj$first = opaque(obj[[2]]) # default: grids/polygons behind, rest front
+		if (missingFirst) # meaning: do plot it
+			first = obj$first
+		if (obj$first == first) {
+			obj$first = NULL
+			do.call(obj[[1]], obj[-1], ...)
+		}
+	} else {
+		sp = sapply(obj, is, "Spatial")
+		stopifnot(any(sp))
+		lapply(obj[sp], function(x) { 
+			if (missingFirst || identical(obj$first, first) || opaque(x) == first)
+				do.call(sppanel, append(x, obj[!sp]), ...)
+		})
+	}
+}
+setMethod("sppanel", "list", sppanelList)
+
 getFormulaLevelplot = function(sdf, zcol) {
 	if (length(zcol) > 1)
 		as.formula(paste("z~", paste(dimnames(coordinates(sdf))[[2]], 
@@ -150,7 +180,8 @@ getFormulaLevelplot = function(sdf, zcol) {
 spplot.grid = function(obj, zcol = names(obj), ..., names.attr, 
 		scales = list(draw = FALSE), xlab = NULL, ylab = NULL, 
 		aspect = mapasp(obj,xlim,ylim), panel = panel.gridplot, sp.layout = NULL, formula, 
-		xlim = bbox(obj)[1,], ylim = bbox(obj)[2,], checkEmptyRC = TRUE) {
+		xlim = bbox(obj)[1,], ylim = bbox(obj)[2,], checkEmptyRC = TRUE,
+		col.regions = get_col_regions()) {
 	if (is.null(zcol)) stop("no names method for object")
 	if (checkEmptyRC)
 		sdf = addNAemptyRowsCols(obj) # returns SpatialPointsDataFrame
@@ -171,12 +202,13 @@ spplot.grid = function(obj, zcol = names(obj), ..., names.attr,
 	scales = longlat.scales(obj, scales, xlim, ylim)
 	args = append(list(formula, data = as(sdf, "data.frame"), 
 		aspect = aspect, panel = panel, xlab = xlab, ylab = ylab, scales = scales,
-		sp.layout = sp.layout, xlim = xlim, ylim = ylim), list(...))
+		sp.layout = sp.layout, xlim = xlim, ylim = ylim, col.regions = col.regions), 
+		list(...))
 	# deal with factor variables:
 	if (all(unlist(lapply(obj@data[zcol], is.factor)))) {
-		if (!is.null(args$col.regions) &&
-				nlevels(obj@data[[zcol[1]]]) != length(args$col.regions))
-			stop("length of col.regions should match number of factor levels")
+		#if (!is.null(args$col.regions) &&
+		#		nlevels(obj@data[[zcol[1]]]) != length(args$col.regions))
+		#	stop("length of col.regions should match number of factor levels")
 		args$data[[zcol2]] = as.numeric(args$data[[zcol2]])
 		if (is.null(args$colorkey) || (is.logical(args$colorkey) && args$colorkey)
 				|| (is.list(args$colorkey) && is.null(args$colorkey$at) && 
@@ -189,7 +221,7 @@ spplot.grid = function(obj, zcol = names(obj), ..., names.attr,
 		} else
 			args = append(args, colorkey.factor(obj[[zcol[1]]], ck, FALSE))
 	}
-	ret = do.call("levelplot", args)
+	ret = do.call(levelplot, args)
     if (exists("panel.levelplot.raster"))
         lattice.options("panel.levelplot" = opan)
 	ret 
@@ -203,7 +235,8 @@ spplot.polygons = function(obj, zcol = names(obj), ..., names.attr,
 		scales = list(draw = FALSE), xlab = NULL, ylab = NULL, 
 		aspect = mapasp(obj,xlim,ylim), 
 		panel = panel.polygonsplot, sp.layout = NULL, formula, 
-		xlim = bbox(obj)[1,], ylim = bbox(obj)[2,]) {
+		xlim = bbox(obj)[1,], ylim = bbox(obj)[2,],
+		col.regions = get_col_regions()) {
 
 	if (is.null(zcol)) stop("no names method for object")
 	sdf = as(obj, "data.frame")
@@ -235,11 +268,12 @@ spplot.polygons = function(obj, zcol = names(obj), ..., names.attr,
 	args = append(list(formula, data = as(sdf, "data.frame"),
 		aspect = aspect, grid.polygons = grid.polygons, panel =
 		panel, xlab = xlab, ylab = ylab, scales = scales,
-		sp.layout = sp.layout, xlim = xlim, ylim = ylim), list(...))
+		sp.layout = sp.layout, xlim = xlim, ylim = ylim,
+		col.regions = col.regions), list(...))
 	if (all(unlist(lapply(obj@data[zcol], is.factor)))) {
-		if (!is.null(args$col.regions) &&
-				nlevels(obj@data[[zcol[1]]]) != length(args$col.regions))
-			stop("length of col.regions should match number of factor levels")
+		#if (!is.null(args$col.regions) &&
+		#		nlevels(obj@data[[zcol[1]]]) != length(args$col.regions))
+		#	stop("length of col.regions should match number of factor levels")
 		args$data[[zcol2]] = as.numeric(args$data[[zcol2]])
 		if (is.null(args$colorkey) 
 				|| (is.logical(args$colorkey) && args$colorkey)
@@ -253,7 +287,7 @@ spplot.polygons = function(obj, zcol = names(obj), ..., names.attr,
 		} else
 			args = append(args, colorkey.factor(obj[[zcol[1]]], ck, FALSE))
 	}
-	do.call("levelplot", args)
+	do.call(levelplot, args)
 }
 
 setMethod("spplot", signature("SpatialPolygonsDataFrame"), spplot.polygons)
@@ -265,7 +299,8 @@ spplot.points = function(obj, zcol = names(obj), ..., names.attr,
 		sp.layout = NULL, identify = FALSE, formula,
 		xlim = bbexpand(bbox(obj)[1,], 0.04), 
 		ylim = bbexpand(bbox(obj)[2,], 0.04),
-		edge.col = "transparent", colorkey = FALSE) 
+		edge.col = "transparent", colorkey = FALSE,
+		col.regions = get_col_regions()) 
 {
 
 	if (is.null(zcol)) stop("no names method for object")
@@ -290,13 +325,14 @@ spplot.points = function(obj, zcol = names(obj), ..., names.attr,
 	args.xyplot = append(list(formula, data = as(sdf, "data.frame"), 
 		panel = panel, aspect = aspect, scales = scales, 
 		xlab = xlab, ylab = ylab, sp.layout = sp.layout,
-		xlim = xlim, ylim = ylim, edge.col = edge.col), dots)
+		xlim = xlim, ylim = ylim, edge.col = edge.col,
+		col.regions = col.regions), dots)
 	z = create.z(as(obj, "data.frame"), zcol)
-	args.xyplot = Fill.call.groups(args.xyplot, z = z, edge.col = edge.col, 
+	args.xyplot = fill.call.groups(args.xyplot, z = z, edge.col = edge.col, 
 		colorkey = colorkey, ...)
 	# debug:
 	#print(args.xyplot)
-	plt = do.call("xyplot", args.xyplot)
+	plt = do.call(xyplot, args.xyplot)
 	if (!(is.logical(identify) && identify==FALSE) && interactive()) {
 		print(plt)
 		if (!(is.numeric(identify) && length(identify) == 2))
@@ -313,6 +349,9 @@ spplot.points = function(obj, zcol = names(obj), ..., names.attr,
 }
 setMethod("spplot", signature("SpatialPointsDataFrame"), spplot.points)
 
+setMethod("spplot", signature("SpatialMultiPointsDataFrame"), 
+	function(obj, ...) spplot.points(as(obj, "SpatialPointsDataFrame"), ...))
+
 create.z = function(df, zcol) {
 	if (is.logical(df[[zcol[1]]])) {
 		z = stack(df[zcol])[[1]]
@@ -328,24 +367,9 @@ create.z = function(df, zcol) {
 }
 
 panel.gridplot = function(x, y, z, subscripts, ..., sp.layout) {
-	# set first = TRUE defaults for polygons objects in sp.layout:
-	if (!missing(sp.layout) && inherits(sp.layout, "list")) {
-		if (inherits(sp.layout[[1]], "list")) {
-			for (i in seq(along = sp.layout)) {
-				if (inherits(sp.layout[[i]], "list")) {
-					sp.i = sp.layout[[i]]
-					if (is.null(sp.i$first) && sp.i[[1]] == "sp.polygons")
-						sp.layout[[i]]$first = TRUE
-				}
-			}
-		} else if (is.null(sp.layout$first) && sp.layout[[1]] == "sp.polygons")
-			sp.layout$first = TRUE
-	}
-	# print(sp.layout)
-
-	sp.panel.layout(sp.layout, panel.number(), first = TRUE)
+	sppanel(list(sp.layout), panel.number(), first = TRUE)
 	panel.levelplot(x, y, z, subscripts, ...)
-	sp.panel.layout(sp.layout, panel.number())
+	sppanel(list(sp.layout), panel.number(), first = FALSE)
 }
 
 panel.polygonsplot =
@@ -368,7 +392,7 @@ function (x, y, z, subscripts, at = pretty(z), shrink, labels = NULL,
    			rep(col.regions, length = numcol)
    		else col.regions[floor(1 + (1:numcol - 1) * (numcol.r - 1)/(numcol - 1))]
 	zcol <- rep(NA, length(z))
-	for (i in seq(along = col.regions)) zcol[!is.na(x) & !is.na(y) & 
+	for (i in seq_along(col.regions)) zcol[!is.na(x) & !is.na(y) & 
       			!is.na(z) & z >= at[i] & z < at[i + 1]] <- i
 	label.style <- match.arg(label.style)
 	x <- as.numeric(x[subscripts])
@@ -376,10 +400,7 @@ function (x, y, z, subscripts, at = pretty(z), shrink, labels = NULL,
 	z <- as.numeric(z[subscripts])
 	zcol <- as.numeric(zcol[subscripts])
 
-	#EJP,2010-10-8: 
-	#if (is(grid.polygons, "SpatialLines"))
-	#	sp.panel.layout(sp.layout, panel.number())
-	sp.panel.layout(sp.layout, panel.number(), first = TRUE)
+	sppanel(list(sp.layout), panel.number(), first = TRUE)
 	if (any(subscripts)) {
 		if (is(grid.polygons, "SpatialLines")) {
 			sp.lines3 = function(x, col, ...) panel.lines(coordinates(x), col = col, ...)
@@ -389,6 +410,7 @@ function (x, y, z, subscripts, at = pretty(z), shrink, labels = NULL,
 		} else {
 			pls = slot(grid.polygons, "polygons")
    			pO = slot(grid.polygons, "plotOrder")
+			col = rep(col, length.out = length(grid.polygons))
    			for (i in pO) {
 				if (get_Polypath()) {
 					obj = as(as(grid.polygons[i,], "SpatialLines"),
@@ -398,7 +420,7 @@ function (x, y, z, subscripts, at = pretty(z), shrink, labels = NULL,
 					fill = col.regions[zcol[i]]
 					alpha = alpha.regions
 					grid.path(cc[,1], cc[,2], id, default.units = "native",
-						gp = gpar(col = col, fill = fill, alpha = alpha, 
+						gp = gpar(col = col[i], fill = fill, alpha = alpha, 
 							lwd = lwd, lty = lty, ...))
 				} else {
        				Srs <- slot(pls[[i]], "Polygons")
@@ -424,19 +446,15 @@ function (x, y, z, subscripts, at = pretty(z), shrink, labels = NULL,
    			}
 		}
 	}
-	# EJP, 2010-10-8
-	#if (!is(grid.polygons, "SpatialLines"))
-	#	sp.panel.layout(sp.layout, panel.number())
-	sp.panel.layout(sp.layout, panel.number())
+	sppanel(list(sp.layout), panel.number(), first = FALSE)
 }
 
 panel.pointsplot = function(sp.layout, x, y, subscripts, groups, col, cex,
 		pch, ...) {
-	sp.panel.layout(sp.layout, panel.number())
-	#panel.superpose(x, y, subscripts, col = col, ...)
+	sppanel(list(sp.layout), panel.number(), first = TRUE)
 	lpoints(x, y, fill = groups[subscripts], col = col[subscripts], 
 		cex = cex[subscripts], pch = pch[subscripts], ...)
-	#panel.xyplot(x, y, pch = 21, fill = fill, ...)
+	sppanel(list(sp.layout), panel.number(), first = FALSE)
 }
 
 SpatialPolygons2Grob = function(obj, fill) {
@@ -540,10 +558,10 @@ sp.theme = function(set = FALSE, regions = list(col = bpy.colors(100)), ...) {
 }
 
 spplot.key = function(sp.layout, rows = 1, cols = 1) {
-	for (i in seq(along=rows)) {
-		for (j in seq(along=cols)) {
+	for (i in seq_along(rows)) {
+		for (j in seq_along(cols)) {
 			trellis.focus("panel", cols[j], rows[i], highlight = FALSE)
-			sp.panel.layout(sp.layout)
+			sppanel(sp.layout)
 			trellis.unfocus()
 		}
 	}
@@ -556,7 +574,7 @@ spplot.key = function(sp.layout, rows = 1, cols = 1) {
 
 longlat.scales = function(obj, scales, xlim, ylim) {
 	isp = is.projected(obj)
-	if (!is.null(scales$draw) && !is.na(isp) && !isp) {
+	if (!is.null(scales$draw) && scales$draw && !is.na(isp) && !isp) {
 		# long lat -- x:
 		if (is.null(scales$x))
 			scales$x = list()
@@ -626,7 +644,7 @@ addNAemptyRowsCols = function(obj) {
 	missing.x = which(apply(missingpatt, 1, all))
 	missing.y = which(apply(missingpatt, 2, all))
 
-	xy = coordinates(obj)[,1:2]
+	xy = coordinates(obj)[,1:2,drop=FALSE]
 	coordvals = coordinatevalues(obj@grid)
 	missing.x = coordvals[[1]][missing.x]
 	missing.y = coordvals[[2]][missing.y]
@@ -644,14 +662,15 @@ addNAemptyRowsCols = function(obj) {
 	obj
 }
 
-Fill.call.groups <-
-function (lst, z, ..., cuts = ifelse(colorkey, 100, 5), 
-	col.regions = trellis.par.get("regions")$col, 
+fill.call.groups <-
+function (lst, z, ..., cuts = ifelse(identical(FALSE, colorkey), 5, 100), 
+	#col.regions = trellis.par.get("regions")$col, 
     legendEntries = "", pch, cex = 1, do.fill = TRUE, do.log = FALSE, 
-    key.space = ifelse(colorkey, "right", "bottom"), 
+    key.space = ifelse(identical(FALSE, colorkey), "bottom", "right"), 
 	cex.key, edge.col, colorkey) 
 {
     dots = list(...)
+	col.regions = lst$col.regions
     if (is.numeric(z)) {
         if (length(cuts) > 1) 
             ncuts = length(cuts) - 1
@@ -719,11 +738,14 @@ function (lst, z, ..., cuts = ifelse(colorkey, 100, 5),
 	if (length(cex) == n) {
 		cex.key = cex
 		lst$cex = cex[groups]
+		lst$cex[is.na(lst$cex)] = 0
 	} else if (missing(cex.key))
 		cex.key = mean(cex, na.rm = TRUE)
 
 	# do key:
-	if (colorkey) {
+	if (is.list(colorkey))
+		lst$legend = colorkey
+	else if (isTRUE(colorkey)) {
 		lst$legend = list(
 			right = list(
 				fun = draw.colorkey,
@@ -782,3 +804,15 @@ function (lst, z, ..., cuts = ifelse(colorkey, 100, 5),
 	}
     return(lst)
 }
+
+panel.RgoogleMaps <- function(map) {
+	bb = bb2merc(map, "RgoogleMaps")
+	grid.raster(map$myTile, mean(bb[1,]), mean(bb[2,]), diff(bb[1,]), diff(bb[2,]), 
+		default.units = "native", interpolate = FALSE)
+}
+
+panel.ggmap <- function(map) {
+	bb = bb2merc(map, "ggmap")
+	grid.raster(map, mean(bb[1,]), mean(bb[2,]), diff(bb[1,]), diff(bb[2,]), 
+		default.units = "native", interpolate = FALSE)
+}
