@@ -74,11 +74,7 @@ void sp_gcdist(double *lon1, double *lon2, double *lat1, double *lat2,
     f = 1.0/298.257223563;     /* WGS-84 ellipsoid flattening factor */
     
     if (fabs(lat1[0] - lat2[0]) < DOUBLE_EPS) {
-        if (fabs(lon1[0] - lon2[0]) < DOUBLE_EPS) {
-            dist[0] = 0.0;
-            return;
-/* Wouter Buytaert bug caught 100211 */
-        } else if (fabs((fabs(lon1[0]) + fabs(lon2[0])) - 360.0) < DOUBLE_EPS) {
+        if (fabs(fmod(lon1[0] - lon2[0], 360.0)) < DOUBLE_EPS) {
             dist[0] = 0.0;
             return;
         }
